/**
 * CMI : Cluster Method Invocation
 * Copyright (C) 2008 Bull S.A.S.
 * Contact: carol@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 * --------------------------------------------------------------------------
 * $Id: JMSClusterViewManager.java 1999 2008-08-06 11:47:15Z loris $
 * --------------------------------------------------------------------------
 */

package org.ow2.cmi.controller.server.impl.jms;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.ObjectStreamClass;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.UUID;
import java.util.concurrent.ConcurrentHashMap;

import javax.jms.BytesMessage;
import javax.jms.Connection;
import javax.jms.ConnectionFactory;
import javax.jms.JMSException;
import javax.jms.Message;
import javax.jms.MessageConsumer;
import javax.jms.MessageListener;
import javax.jms.MessageProducer;
import javax.jms.Session;
import javax.jms.TextMessage;
import javax.jms.Topic;
import javax.management.MBeanServerNotification;
import javax.management.Notification;
import javax.management.NotificationListener;
import javax.management.ObjectName;
import javax.naming.Context;
import javax.naming.InitialContext;
import javax.naming.NamingException;

import net.jcip.annotations.ThreadSafe;

import org.ow2.cmi.admin.MBeanUtils;
import org.ow2.cmi.controller.provider.ClusterViewProvider;
import org.ow2.cmi.controller.provider.ServerClusterViewProvider;
import org.ow2.cmi.controller.server.AbsServerClusterViewManager;
import org.ow2.cmi.controller.server.ClusterView;
import org.ow2.cmi.controller.server.DistributedObjectInfo;
import org.ow2.cmi.controller.server.ServerClusterViewManagerException;
import org.ow2.cmi.controller.server.ServerView;
import org.ow2.cmi.controller.server.filter.IFilter;
import org.ow2.cmi.lb.data.PolicyData;
import org.ow2.cmi.reference.CMIReference;
import org.ow2.cmi.reference.ObjectNotFoundException;
import org.ow2.cmi.reference.ServerId;
import org.ow2.cmi.reference.ServerNotFoundException;
import org.ow2.cmi.reference.ServerRef;
import org.ow2.util.log.Log;
import org.ow2.util.log.LogFactory;

/**
 * A JMS-based server cluster view manager.
 * @author Loris Bouzonnet
 */
@ThreadSafe
public final class JMSClusterViewManager extends AbsServerClusterViewManager
implements MessageListener, NotificationListener {

    /**
     * Logger.
     */
    private static final Log logger = LogFactory
            .getLog(JMSClusterViewManager.class);

    /**
     * Topic of Dad (since hierarchical topic).
     */
    private Topic superTopic;


    private Connection superTopicConnection;

    private Session superSession;

    /**
     * To send message for Dad (since hierarchical topic).
     */
    private MessageProducer superPublisher;

    /**
     * Topic of my twins and me (since clustered topic).
     */
    private Topic ourTopic;

    /**
     * To receive message of our children.
     */
    private MessageConsumer ourSubscriber;

    private MessageProducer ourPublisher;

    private Connection ourTopicConnection;

    private Session ourSessionSubscriber;

    private Session ourSessionPublisher;

    private ClusterView clusterView = new ClusterView();

    private ObjectName delegate;

    private ObjectName objectNameToListen;

    private boolean initialized = false;

    private ClusterViewCleaner clusterViewCleaner;

    private Heartbeat heartbeat;

    private boolean registrationOpened;

    @Override
    public void doStart() {
        if(getObjectNameToListen() != null) {
            try {
                objectNameToListen = new ObjectName(getObjectNameToListen());
            } catch (Exception e) {
                logger.error("Unable to check if JMS is available", e);
                throw new ServerClusterViewManagerException(
                        "Unable to check if JMS is available", e);
            }
        }
        if(getObjectNameToListen() == null
                || MBeanUtils.getMBeanServer().isRegistered(objectNameToListen)) {
            registrationOpened = true;
        } else {
            // Register a listener to be notified of the availability of JMS
            try {
                delegate = new ObjectName("JMImplementation:type=MBeanServerDelegate");
                MBeanUtils.getMBeanServer().addNotificationListener(
                        delegate, this, null, null);
            } catch (Exception e) {
                logger.error("Unable to add the listener for JMS", e);
                throw new ServerClusterViewManagerException(
                        "Unable to add the listener for JMS", e);
            }
            logger.info("CMI server waiting for a JMS implementation...");
        }
    }


    @Override
    public void doStop() {
        registrationOpened = false;
        setImplementationAvailable(false);
        if(delegate != null) {
            if(MBeanUtils.getMBeanServer().isRegistered(delegate)) {
                try {
                    MBeanUtils.getMBeanServer().removeNotificationListener(
                            delegate, this);
                } catch (Exception e) {
                    logger.warn("Unable to remove the listener for JMS", e);
                }
            }
        }
        try {
            if(superTopic != null) {
                superPublisher.close();
                superSession.close();
                superTopicConnection.close();
            }
            if(ourPublisher != null) {
                ourPublisher.close();
                ourSessionPublisher.close();
            }
            if(ourSubscriber != null) {
                ourSubscriber.close();
                ourSessionSubscriber.close();
            }
            if(ourTopicConnection != null) {
                ourTopicConnection.close();
            }
        } catch (JMSException e) {
            logger.warn("Error while closing the JMS objects", e);
        }
        clusterView.clear();
    }


    public boolean isRegistrationOpened() {
        return registrationOpened;
    }

    public void handleNotification(final Notification notification, final Object handback) {
        if (notification.getType().equals(
                MBeanServerNotification.REGISTRATION_NOTIFICATION)) {
            if(objectNameToListen.equals(((MBeanServerNotification) notification).getMBeanName())) {
                registrationOpened = true;
                if(getState().equals(State.STARTED)) {
                    try {
                        new InitialContext();
                    } catch (NamingException e) {
                        logger.warn("Unable to create a cmi context", e);
                    }
                }
            }
        } else if(notification.getType().equals(
                MBeanServerNotification.UNREGISTRATION_NOTIFICATION)) {
            if(objectNameToListen.equals(((MBeanServerNotification) notification).getMBeanName())) {
                logger.warn("JMS no more available: auto-stopping the CMI server.");
                ourSessionPublisher = null;
                ourSessionSubscriber = null;
                ourPublisher = null;
                ourSubscriber = null;
                ourTopic = null;
                ourTopicConnection = null;
                superSession = null;
                superPublisher = null;
                superTopic = null;
                superTopicConnection = null;
                stop();
            }
        }
    }

    @Override
    protected void initServerConfig(final String contextFactory,
            final ServerId serverId, final Context cmiContext) {

        // TODO Multi-protocol support
        if(!initialized) {

            if(getSuperTopicName() != null) {
                initSuperTopic(contextFactory);
            }
            if(getOurTopicName() != null) {
                initOurTopic(contextFactory, serverId);
            }
            super.initServerConfig(contextFactory, serverId, cmiContext);

            if(ourTopic == null && superTopic == null) {
                logger.error("No topic available!");
                throw new ServerClusterViewManagerException("No topic available!");
            }

            setImplementationAvailable(true);

            logger.info("CMI server available.");

            initHeartBeat();
            if(ourTopic != null || superTopic == null) {
                initClusterViewCleaner();
            }

            initialized = true;
        }
    }

    private void initClusterViewCleaner() {
        clusterViewCleaner = new ClusterViewCleaner();
        getCmiThreadFactory().newThread(clusterViewCleaner, "Cluster View Cleaner").start();
    }


    private void initHeartBeat() {
        heartbeat = new Heartbeat();
        getCmiThreadFactory().newThread(heartbeat, "Heart Beat").start();
    }


    private void initSuperTopic(final String contextFactory)
    throws ServerClusterViewManagerException {
        Hashtable<String, String> env = new Hashtable<String, String>();
        env.put(Context.INITIAL_CONTEXT_FACTORY, contextFactory);
        env.put(Context.PROVIDER_URL, getSuperProviderURL());
        Context context;
        try {
            context = new InitialContext(env);
        } catch (NamingException e) {
            logger.error("Unable to create a context", e);
            throw new ServerClusterViewManagerException("Unable to create a context", e);
        }
        try {
            superTopic = (Topic) context.lookup(getSuperTopicName());

            // Start the connection
            ConnectionFactory tcf = (ConnectionFactory) context.lookup(getConnectionFactoryName());
            superTopicConnection = tcf.createConnection();
            superSession =
                superTopicConnection.createSession(false, Session.AUTO_ACKNOWLEDGE);
            superPublisher = superSession.createProducer(superTopic);
            superTopicConnection.start();
        } catch (Exception e) {
            logger.error("Unable to initialize the super topic", e);
            throw new ServerClusterViewManagerException("Unable to initialize the super topic", e);
        } finally {
            try {
                context.close();
            } catch (NamingException e) {
                logger.error("Unable to close the context", e);
            }
        }
    }

    private void initOurTopic(final String contextFactory, final ServerId serverId) {
        Hashtable<String, String> env = new Hashtable<String, String>();
        env.put(Context.INITIAL_CONTEXT_FACTORY, contextFactory);
        String providerURL = getOurProviderURL();
        if(providerURL == null) {
            providerURL = serverId.getProviderURL();
        }
        env.put(Context.PROVIDER_URL, providerURL);
        Context context;
        try {
            context = new InitialContext(env);
        } catch (NamingException e) {
            logger.error("Unable to create a context", e);
            throw new ServerClusterViewManagerException("Unable to create a context", e);
        }
        // Try to retrieve the topic
        try {
            ourTopic = (Topic) context.lookup(getOurTopicName());
        } catch (NamingException e) {
            if(getOurProviderURL() != null) {
                logger.error("No topic found with name {0} for the provider URL {1}",
                        getOurTopicName(), getOurProviderURL(), e);
                throw new ServerClusterViewManagerException(
                        "No topic found with name " + getOurTopicName()
                        + " for the provider URL " + getOurProviderURL());
            }
            logger.debug("No topic found with name {0} in the local registry.", getOurTopicName(), e);
            ourTopic = null;
        }

        try {
            if(ourTopic == null) {
                createOurTopic();
                context.rebind(getOurTopicName(), ourTopic);
            }

            // Try to retrieve an existing cluster view
            if(getClusterViewProviderURL() == null) {
                logger.warn("Cannot retrieve the cluster view: missing infos about its provider.");
            } else {
                Context contextForProvider = null;
                try {
                    Hashtable<String, String> envForProvider = new Hashtable<String, String>();
                    envForProvider.put(Context.INITIAL_CONTEXT_FACTORY, contextFactory);
                    envForProvider.put(Context.PROVIDER_URL, getClusterViewProviderURL());
                    contextForProvider = new InitialContext(envForProvider);
                    clusterView = ((ServerClusterViewProvider)
                            contextForProvider.lookup(getJmsProviderName())).getClusterView();
                } catch (NamingException e) {
                    logger.warn("Cannot retrieve the cluster view: no provider found for the jndi name {0}",
                            getJmsProviderName(), e);
                } finally {
                    if(contextForProvider != null) {
                        try {
                            contextForProvider.close();
                        } catch (NamingException e) {
                            logger.error("Unable to close the context", e);
                        }
                    }
                }
            }

            if(isClusterViewShared()) {
                try {
                    ClusterViewProvider jmsProvider = new JmsProvider(this);
                    context.rebind(getJmsProviderName(), jmsProvider);
                } catch(Exception e) {
                    logger.warn("Cannot share the cluster view provider with name " + getJmsProviderName(), e);
                }
            }

            // Start the connection
            ConnectionFactory tcf = (ConnectionFactory) context.lookup(getConnectionFactoryName());
            ourTopicConnection = tcf.createConnection();
            ourSessionSubscriber = ourTopicConnection.createSession(false, Session.AUTO_ACKNOWLEDGE);
            ourSubscriber = ourSessionSubscriber.createConsumer(ourTopic);
            ourSubscriber.setMessageListener(this);
            if(isPublisher()) {
                ourSessionPublisher = ourTopicConnection.createSession(false, Session.AUTO_ACKNOWLEDGE);
                ourPublisher = ourSessionPublisher.createProducer(ourTopic);
            }
            ourTopicConnection.start();
        } catch (Exception e) {
            logger.error("Unable to initialize our topic", e);
            throw new ServerClusterViewManagerException("Unable to initialize our topic", e);
        } finally {
            try {
                context.close();
            } catch (NamingException e) {
                logger.error("Unable to close the context", e);
            }
        }
    }

    private void createOurTopic() {
        if(getOurTopicName() != null && ourTopic == null) {
            try {
                ourTopic = JORAMHelper.createTopic(getOurTopicName(), superTopic);
            } catch (Exception e) {
                logger.error("Cannot create the topic {0}", getOurTopicName(), e);
                throw new ServerClusterViewManagerException("Cannot create the topic " + getOurTopicName(), e);
            }
        }
    }

    @Override
    protected boolean containObject(final String objectName) {
        return clusterView.containsObject(objectName);
    }

    @Override
    protected DistributedObjectInfo getDistributedObjectInfo(final String objectName)
            throws ObjectNotFoundException {
        return clusterView.getObjectInfo(objectName);
    }

    @Override
    public void broadCastArchive(final Object archiveId, final byte[] bytesOfFile) {
        throw new UnsupportedOperationException("TODO");
    }

    @Override
    protected void initStats() {
        // TODO
    }

    @Override
    public boolean isAlreadyDistributed(final Object archiveId) {
        throw new UnsupportedOperationException("TODO");
    }

    @Override
    public void removeDistributedArchive(final Object archiveId) {
        throw new UnsupportedOperationException("TODO");
    }

    private void sendMessage(final Object data, final Map<String, Serializable> properties) {
        if(!(data instanceof Serializable)) {
            throw new IllegalArgumentException("The provided object is not serializable.");
        }
        Session session = null;
        MessageProducer publisher = null;
        if(ourSessionPublisher != null) {
            session = ourSessionPublisher;
            publisher = ourPublisher;
        } else if(superSession != null) {
            session = superSession;
            publisher = superPublisher;
        }
        if(session != null) {
            try {
                BytesMessage msg = session.createBytesMessage();
                msg.writeBytes(serialize(data));
                for(String key : properties.keySet()) {
                    msg.setObjectProperty(key, properties.get(key));
                }
                publisher.send(msg);
            } catch (Exception e) {
                logger.error("Unable to send a message to add data", e);
            }
        }
    }

    @Override
    protected void addDistributedObjectInfo(final String objectName,
            final DistributedObjectInfo distributedObjectInfo) {
        addLocalDistributedObjectInfo(objectName, distributedObjectInfo);
        HashMap<String, Serializable> properties = new HashMap<String, Serializable>();
        properties.put("action", "add");
        sendMessage(distributedObjectInfo, properties);
    }


    /**
     * Sets informations on the clustered object with the given name.
     *
     * @param objectName
     *            a name of object
     * @param distributedObjectInfo
     *            informations on the clustered object
     */
    @Override
    protected void setDistributedObjectInfo(final String objectName,
            final DistributedObjectInfo distributedObjectInfo) {
        setLocalDistributedObjectInfo(objectName, distributedObjectInfo);
        HashMap<String, Serializable> properties = new HashMap<String, Serializable>();
        properties.put("action", "set");
        sendMessage(distributedObjectInfo, properties);
    }

    public void addCMIReference(final CMIReference cmiReference) {
        addLocalCMIReference(cmiReference);
        HashMap<String, Serializable> properties = new HashMap<String, Serializable>();
        properties.put("action", "add");
        sendMessage(cmiReference, properties);
    }


    public List<CMIReference> getCMIReferences(final String objectName)
            throws ObjectNotFoundException {
        return clusterView.getCMIReferences(objectName);
    }

    public Collection<CMIReference> getCMIReferences(final String objectName,
            final String protocolName)
            throws ObjectNotFoundException {
        return clusterView.getCMIReferences(objectName, protocolName);
    }

    public Set<String> getObjectNames() {
        return clusterView.getObjectNames();
    }


    private void setLocalDistributedObjectInfo(final String objectName,
            final DistributedObjectInfo distributedObjectInfo) {
        DistributedObjectInfo oldDistributedObjectInfo =
            clusterView.addObjectInfo(distributedObjectInfo, true);
        if(!oldDistributedObjectInfo.equals(distributedObjectInfo)) {
            PolicyData policyData = oldDistributedObjectInfo.getPolicyData();
            if(!policyData.equals(distributedObjectInfo.getPolicyData())) {
                try {
                    updatePolicy(objectName);
                } catch (Exception e) {
                    logger.error("Cannot update the policy for the object with name " + objectName, e);
                }
            }
            try {
               updatePool(oldDistributedObjectInfo, distributedObjectInfo);
            } catch (ObjectNotFoundException e) {
                logger.error("Cannot update the pool configuration for the object with name " + objectName, e);
            }
        }
    }

    private void addLocalDistributedObjectInfo(final String objectName,
            final DistributedObjectInfo distributedObjectInfo) {
        clusterView.addObjectInfo(distributedObjectInfo, false);
        try {
            updatePolicy(objectName);
        } catch (Exception e) {
            logger.error("Cannot update the policy for the object with name " + objectName, e);
        }
        try {
            updatePool(null, distributedObjectInfo);
        } catch (ObjectNotFoundException e) {
            logger.error("Cannot update the pool configuration for the object with name " + objectName, e);
        }
    }

    private void addLocalCMIReference(final CMIReference cmiReference) {
        clusterView.addCMIReference(cmiReference);
    }

    public void removeCMIReference(final CMIReference cmiReference) {
        try {
            removeLocalCMIReference(cmiReference);
            HashMap<String, Serializable> properties = new HashMap<String, Serializable>();
            properties.put("action", "rem");
            sendMessage(cmiReference, properties);
        } catch (ObjectNotFoundException e) {
            logger.debug("Cannot remove the CMIReference {0}", cmiReference, e);
        }
    }

    private void removeLocalCMIReference(final CMIReference cmiReference) throws ObjectNotFoundException {
        clusterView.removeCMIReference(cmiReference);
    }

    public void setDelayToRefresh(final int delay) {
        setLocalDelay(delay);
        HashMap<String, Serializable> properties = new HashMap<String, Serializable>();
        properties.put("delay", "set");
        sendMessage(delay, properties);
    }


    private void setLocalDelay(final int delay) {
        clusterView.setDelay(delay);
    }


    public int getDelayToRefresh() {
        return clusterView.getDelay();
    }

    public List<IFilter> getGlobalFilters() {
        return clusterView.getFilters();
    }

    public void addGlobalFilter(final IFilter filter) throws IllegalArgumentException {
        addFilter(filter);
        HashMap<String, Serializable> properties = new HashMap<String, Serializable>();
        properties.put("action", "add");
        sendMessage(filter, properties);
    }

    public void removeGlobalFilter(final IFilter filter) {
        if(removeFilter(filter)) {
            HashMap<String, Serializable> properties = new HashMap<String, Serializable>();
            properties.put("action", "rem");
            sendMessage(filter, properties);
        }
    }

    private void addFilter(final IFilter filter) {
        clusterView.addFilter(filter);
    }

    private boolean removeFilter(final IFilter filter) {
        return clusterView.removeFilter(filter);
    }


    public boolean isServerBlackListed(final ServerRef serverRef) {
        return clusterView.isServerBlackListed(serverRef);
    }

    public void addServerToBlackList(final ServerRef serverRef) {
        if(!clusterView.addServerToBlackList(serverRef)) {
            HashMap<String, Serializable> properties = new HashMap<String, Serializable>();
            properties.put("black", true);
            sendMessage(serverRef, properties);
        }
    }

    public void removeServerFromBlackList(final ServerRef serverRef) {
        try {
            if(clusterView.removeServerFromBlackList(serverRef)) {
                HashMap<String, Serializable> properties = new HashMap<String, Serializable>();
                properties.put("black", false);
                sendMessage(serverRef, properties);
            }
        } catch (ServerNotFoundException e) {
            logger.debug("Unable to remove {0} from the blacklist", serverRef, e);
        }
    }

    public Set<String> getClusterNames() {
        return clusterView.getClusterNames();
    }


    public int getNbClientsConnectedToProvider() {
        return -1;
    }

    public Set<String> getObjectNames(final String clusterName) {
        return clusterView.getObjectNames(clusterName);
    }

    public void registerClient(final UUID uuid) {
        // TODO
    }


    public boolean isPoolToEmpty(final String objectName)
            throws ObjectNotFoundException {
        throw new UnsupportedOperationException("TODO");
    }

    public void addPoolToEmpty(final String objectName) {
        throw new UnsupportedOperationException("TODO");
    }

    public void removePoolToEmpty(final String objectName) {
        throw new UnsupportedOperationException("TODO");
    }

    public int getLoadFactor(final ServerRef serverRef)
            throws ServerNotFoundException {
        return clusterView.getLoadFactor(serverRef);
    }

    public void setLoadFactor(final ServerRef serverRef, final int loadFactor) {
        if(clusterView.setLoadFactor(serverRef, loadFactor) != loadFactor) {
            HashMap<String, Serializable> properties = new HashMap<String, Serializable>();
            properties.put("load", loadFactor);
            sendMessage(serverRef, properties);
        }
    }

    public long getDateOfConfiguration() {
        // TODO
        return -1;
    }

    public void onMessage(final Message msg) {
        try {
            if(msg instanceof BytesMessage) {
                BytesMessage bytesMessage = (BytesMessage) msg;
                Object object;
                try {
                    object = unserialize(bytesMessage);
                } catch (Exception e) {
                    logger.error("Cannot unserialize the bytes message {0}", bytesMessage, e);
                    return;
                }
                if(object instanceof CMIReference) {
                    CMIReference cmiRef = (CMIReference) object;
                    if(bytesMessage.getObjectProperty("action").equals("add")) {
                        logger.debug("A new object is available at this location: {0}", cmiRef);
                        addLocalCMIReference(cmiRef);
                    } else {
                        logger.debug("A new object is to remove for this location: {0}", cmiRef);
                        try {
                            removeLocalCMIReference(cmiRef);
                        } catch (ObjectNotFoundException e) {
                            logger.debug("Cannot remove the CMIReference {0}", cmiRef, e);
                        }
                    }
                } else if(object instanceof DistributedObjectInfo) {
                    DistributedObjectInfo distributedObjectInfo = (DistributedObjectInfo) object;
                    if(bytesMessage.getObjectProperty("action").equals("add")) {
                        addLocalDistributedObjectInfo(distributedObjectInfo.getObjectName(), distributedObjectInfo);
                    } else {
                        setLocalDistributedObjectInfo(distributedObjectInfo.getObjectName(), distributedObjectInfo);
                    }
                } else if(object instanceof IFilter) {
                    IFilter filter = (IFilter) object;
                    if(bytesMessage.getObjectProperty("action").equals("add")) {
                        addFilter(filter);
                    } else {
                        removeFilter(filter);
                    }
                } else if(object instanceof ServerRef) {
                    ServerRef serverRef = (ServerRef) object;
                    Object value = bytesMessage.getObjectProperty("black");
                    if(value != null && value instanceof Boolean) {
                        if((Boolean) value) {
                            addServerToBlackList(serverRef);
                        } else {
                            removeServerFromBlackList(serverRef);
                        }
                    }
                    value = bytesMessage.getObjectProperty("load");
                    if(value != null && value instanceof Integer) {
                        setLoadFactor(serverRef, (Integer) value);
                    }
                } else if(object instanceof Integer) {
                    Object value = bytesMessage.getObjectProperty("delay");
                    if(value != null && value.equals("set")) {
                        setLocalDelay((Integer) object);
                    }
                }
            }
        } catch (JMSException e) {
            logger.error("Cannot read the message {0}", msg, e);
        }
    }

    private byte[] serialize(final Object object) throws IOException {
        ByteArrayOutputStream bOutputStream = new ByteArrayOutputStream();
        ObjectOutputStream objectOutputStream = new ObjectOutputStream(bOutputStream);
        objectOutputStream.writeObject(object);
        objectOutputStream.flush();
        return bOutputStream.toByteArray();
    }

    private Object unserialize(final BytesMessage bytesMessage)
    throws IOException, JMSException, ClassNotFoundException {
        byte[] bytes = new byte[(int) bytesMessage.getBodyLength()];
        bytesMessage.readBytes(bytes);
        ObjectInputStream stream = new ObjectInputStream(
                new ByteArrayInputStream(bytes)) {
            @Override
            protected Class<?> resolveClass(final ObjectStreamClass desc) throws IOException, ClassNotFoundException {
                String name = desc.getName();
                return Class.forName(name, false, getClass().getClassLoader());
            }
        };
        try {
            return stream.readObject();
        } finally {
            stream.close();
        }
    }

    private String getSuperTopicName() {
        return ((Config) getConfig()).getSuperTopicName();
    }

    private String getOurTopicName() {
        return ((Config) getConfig()).getOurTopicName();
    }

    private String getSuperProviderURL() {
        return ((Config) getConfig()).getSuperProviderURL();
    }

    private String getOurProviderURL() {
        return ((Config) getConfig()).getOurProviderURL();
    }

    private boolean isPublisher() {
        return ((Config) getConfig()).isPublisher();
    }

    private String getClusterViewProviderURL() {
        return ((Config) getConfig()).getClusterViewProviderURL();
    }

    private boolean isClusterViewShared() {
        return ((Config) getConfig()).isClusterViewShared();
    }

    private String getObjectNameToListen() {
        return ((Config) getConfig()).getObjectNameToListen();
    }

    private String getJmsProviderName() {
        return ((Config) getConfig()).getJmsProviderName();
    }

    private String getConnectionFactoryName() {
        return ((Config) getConfig()).getConnectionFactoryName();
    }

    public long getPeriodToBeat() {
        return ((Config) getConfig()).getPeriodToBeat();
    }

    private long getPeriodToCleanUp() {
        return ((Config) getConfig()).getPeriodToCleanUp();
    }

    private long getWaitingBeatNumber() {
        return ((Config) getConfig()).getWaitingBeatNumber();
    }

    public ClusterView getClusterView() {
        return clusterView;
    }

    private class Heartbeat implements Runnable {

        public void run() {
            Connection connection = null;
            Topic topic;
            if(ourTopic != null) {
                connection = ourTopicConnection;
                topic = ourTopic;
            } else if(superTopic != null) {
                connection = superTopicConnection;
                topic = superTopic;
            } else {
                logger.error("No connection available!");
                throw new ServerClusterViewManagerException("No connection available!");
            }
            Session session;
            MessageProducer producer;
            Set<ServerId> registries = getRefOnLocalRegistries();
            List<Message> messages = new ArrayList<Message>(registries.size());
            try {
                session = connection.createSession(false, Session.AUTO_ACKNOWLEDGE);
                producer = session.createProducer(topic);
                for(ServerId serverId : registries) {
                    messages.add(session.createTextMessage(serverId.getProviderURL()));
                }
            } catch (JMSException e) {
                logger.error("Cannot init the producer for the heartbeat!", e);
                throw new ServerClusterViewManagerException(
                        "Cannot init the producer for the heartbeat!", e);
            }
            while(isImplementationAvailable()) {
                try {
                    for(Message msg : messages) {
                        producer.send(msg);
                    }
                } catch (JMSException e) {
                    logger.warn("Cannot send a heartbeat!", e);
                }
                try {
                    Thread.sleep(getPeriodToBeat());
                } catch (InterruptedException e) {
                    // Restore interrupted status
                    Thread.currentThread().interrupt();
                }
            }
        }

    }

    private class ClusterViewCleaner implements Runnable, MessageListener {

        private Map<String, Long> timestamps;


        public void run() {
            timestamps = new ConcurrentHashMap<String, Long>();
            Connection connection = null;
            Topic topic;
            if(ourTopic != null) {
                connection = ourTopicConnection;
                topic = ourTopic;
            } else if(superTopic != null) {
                connection = superTopicConnection;
                topic = superTopic;
            } else {
                logger.error("No connection available!");
                throw new ServerClusterViewManagerException("No connection available!");
            }
            try {
                Session session = connection.createSession(false, Session.AUTO_ACKNOWLEDGE);
                MessageConsumer consumer = session.createConsumer(topic);
                consumer.setMessageListener(this);
            } catch (JMSException e) {
                logger.error("Cannot init the consumer for the cluster view cleaner!", e);
                throw new ServerClusterViewManagerException(
                        "Cannot init the consumer for the cluster view cleaner!", e);
            }
            while(isImplementationAvailable()) {
                try {
                    Thread.sleep(getPeriodToCleanUp());
                } catch (InterruptedException e) {
                    // Restore interrupted status
                    Thread.currentThread().interrupt();
                }
                logger.debug("Cleaning the cluster view...");
                cleanClusterView();
            }
        }

        private boolean isLocalRegistry(final String providerURL) {
            for(ServerId serverId : getRefOnLocalRegistries()) {
                if(serverId.getProviderURL().equals(providerURL)) {
                    return true;
                }
            }
            return false;
        }

        private synchronized void cleanClusterView() {
            Map<String, ? extends ServerView> serverViews = clusterView.getServerViews();
            synchronized (serverViews) {
                Iterator<String> it = serverViews.keySet().iterator();
                while(it.hasNext()) {
                    String providerURL = it.next();
                    // Get the last timestamp
                    Long timestamp = timestamps.get(providerURL);
                    if(timestamp != null &&
                            timestamp + getPeriodToBeat() * getWaitingBeatNumber() < System.currentTimeMillis()) {
                        if(isLocalRegistry(providerURL)) {
                            logger.warn("The local registry doesn't anymore send heartbeats!", providerURL);
                        } else {
                            logger.debug("Removing {0}", providerURL);
                            clusterView.removeClusteredObjects(providerURL);
                            // Remove its server view
                            it.remove();
                        }
                    }
                }
            }
        }

        public void onMessage(final Message msg) {
            if(msg instanceof TextMessage) {
                try {
                    timestamps.put(((TextMessage) msg).getText(), msg.getJMSTimestamp());
                } catch (JMSException e) {
                    logger.warn("Cannot extract the content of {0}", msg, e);
                }
            }
        }

    }

}
